# voom_mode_fmr1.py
# Last Modified: 2013-10-31
# VOoM -- Vim two-pane outliner, plugin for Python-enabled Vim 7.x
# Website: http://www.vim.org/scripts/script.php?script_id=2657
# Author: Vlad Irnov (vlad DOT irnov AT gmail DOT com)
# License: CC0, see http://creativecommons.org/publicdomain/zero/1.0/

"""
This mode changes absolutely nothing, it is identical to the default mode.
See |voom-mode-fmr|, ../../doc/voom.txt#*voom-mode-fmr*
"""

# Define this mode as an 'fmr' mode.
MTYPE = 0
